@@ -1,21 +1,4 @@
 #ifndef ASMINLINE
-#endif
-/* Copyright (C) 2000  The PARI group.
-
-This file is part of the PARI/GP package.
-
-PARI/GP is free software; you can redistribute it and/or modify it under the
-terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
-ANY WARRANTY WHATSOEVER.
-
-Check the License for details. You should have received a copy of it, along
-with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
-
-/*
-NOASM addll mulll bfffo divll
-*/
 #line 2 "../src/kernel/none/addll.h"
 /* Copyright (C) 2003  The PARI group.
 
@@ -23,7 +6,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -48,7 +32,7 @@ extern long subllx(ulong x, ulong y);
 
 #if defined(__GNUC__) && !defined(DISABLE_INLINE)
 #undef LOCAL_OVERFLOW
-#define LOCAL_OVERFLOW register ulong overflow
+#define LOCAL_OVERFLOW ulong overflow
 
 #define addll(a, b)                                             \
 __extension__ ({                                                \
@@ -126,7 +110,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -161,7 +146,7 @@ extern long addmul(ulong x, ulong y);
 
 #if defined(__GNUC__) && !defined(DISABLE_INLINE)
 #undef LOCAL_HIREMAINDER
-#define LOCAL_HIREMAINDER register ulong hiremainder
+#define LOCAL_HIREMAINDER ulong hiremainder
 
 #define mulll(x, y) \
 __extension__ ({ \
@@ -266,7 +251,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -330,7 +316,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -360,7 +347,7 @@ extern ulong hiremainder;
 
 #if defined(__GNUC__) && !defined(DISABLE_INLINE)
 #undef LOCAL_HIREMAINDER
-#define LOCAL_HIREMAINDER register ulong hiremainder
+#define LOCAL_HIREMAINDER ulong hiremainder
 
 #define divll(n0, d)                                                    \
 __extension__ ({                                                        \
@@ -480,6 +467,142 @@ divll(ulong n0, ulong d)
 
 #endif /* __GNUC__ */
 
+#endif
+#endif
+#line 2 "../src/kernel/ix86/asm0.h"
+/* Copyright (C) 2000  The PARI group.
+
+This file is part of the PARI/GP package.
+
+PARI/GP is free software; you can redistribute it and/or modify it under the
+terms of the GNU General Public License as published by the Free Software
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
+ANY WARRANTY WHATSOEVER.
+
+Check the License for details. You should have received a copy of it, along
+with the package; see the file 'COPYING'. If not, write to the Free Software
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
+
+/* This file defines some "level 0" kernel functions for Intel ix86  */
+/* It is intended for use with an external "asm" definition          */
+
+/*
+ASM addll mulll bfffo divll
+*/
+#ifdef ASMINLINE
+/* Written by Bruno Haible, 1996-1998.
+   addllx8/subllx8 by Bill Allombert, 2011.
+*/
+
+/* This file can assume the GNU C extensions.
+   (It is included only if __GNUC__ is defined.) */
+
+/* Use local variables whenever possible. */
+#define LOCAL_HIREMAINDER  ulong hiremainder
+#define LOCAL_OVERFLOW     ulong overflow
+
+#define addll(a,b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
+   __asm__ ("addl %3,%0 ; adcl %1,%1" \
+        : "=r" (__value), "=r" (overflow) \
+        : "0" (__arg1), "g" (__arg2), "1" ((ulong)0) \
+        : "cc"); \
+  __value; \
+})
+
+#define addllx(a,b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b), __temp; \
+   __asm__ ("subl %5,%2 ; adcl %4,%0 ; adcl %1,%1" \
+        : "=r" (__value), "=&r" (overflow), "=&r" (__temp) \
+        : "0" (__arg1), "g" (__arg2), "g" (overflow), "1" ((ulong)0), "2" ((ulong)0) \
+        : "cc"); \
+  __value; \
+})
+
+#define addllx8(a,b,c,overflow) \
+do { long *__arg1 = a, *__arg2 = b, *__out = c; \
+     ulong __temp; \
+   __asm__ ("subl %5, %0 \n\t" \
+            "movl    (%2), %0 ; adcl    (%3),%0; movl %0,    (%4) \n\t" \
+            "movl  -4(%2), %0 ; adcl  -4(%3),%0; movl %0,  -4(%4) \n\t" \
+            "movl  -8(%2), %0 ; adcl  -8(%3),%0; movl %0,  -8(%4) \n\t" \
+            "movl -12(%2), %0 ; adcl -12(%3),%0; movl %0, -12(%4) \n\t" \
+            "movl -16(%2), %0 ; adcl -16(%3),%0; movl %0, -16(%4) \n\t" \
+            "movl -20(%2), %0 ; adcl -20(%3),%0; movl %0, -20(%4) \n\t" \
+            "movl -24(%2), %0 ; adcl -24(%3),%0; movl %0, -24(%4) \n\t" \
+            "movl -28(%2), %0 ; adcl -28(%3),%0; movl %0, -28(%4) \n\t" \
+            "adcl  %1, %1" \
+        : "=&r" (__temp), "=&r" (overflow) \
+        : "r" (__arg1), "r" (__arg2), "r" (__out), "g" (overflow), "0" ((ulong)0), "1" ((ulong)0)        : "cc"); \
+} while(0)
+
+#define subll(a,b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
+   __asm__ ("subl %3,%0 ; adcl %1,%1" \
+        : "=r" (__value), "=r" (overflow) \
+        : "0" (__arg1), "g" (__arg2), "1" ((ulong)0) \
+        : "cc"); \
+  __value; \
+})
+
+#define subllx(a,b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b), __temp; \
+   __asm__ ("subl %5,%2 ; sbbl %4,%0 ; adcl %1,%1" \
+        : "=r" (__value), "=&r" (overflow), "=&r" (__temp) \
+        : "0" (__arg1), "g" (__arg2), "g" (overflow), "1" ((ulong)0), "2" ((ulong)0) \
+        : "cc"); \
+  __value; \
+})
+
+#define subllx8(a,b,c,overflow) \
+do { long *__arg1 = a, *__arg2 = b, *__out = c; \
+     ulong __temp; \
+   __asm__ ("subl %5, %0 \n\t" \
+            "movl    (%2), %0 ; sbbl    (%3),%0; movl %0,    (%4) \n\t" \
+            "movl  -4(%2), %0 ; sbbl  -4(%3),%0; movl %0,  -4(%4) \n\t" \
+            "movl  -8(%2), %0 ; sbbl  -8(%3),%0; movl %0,  -8(%4) \n\t" \
+            "movl -12(%2), %0 ; sbbl -12(%3),%0; movl %0, -12(%4) \n\t" \
+            "movl -16(%2), %0 ; sbbl -16(%3),%0; movl %0, -16(%4) \n\t" \
+            "movl -20(%2), %0 ; sbbl -20(%3),%0; movl %0, -20(%4) \n\t" \
+            "movl -24(%2), %0 ; sbbl -24(%3),%0; movl %0, -24(%4) \n\t" \
+            "movl -28(%2), %0 ; sbbl -28(%3),%0; movl %0, -28(%4) \n\t" \
+            "adcl  %1, %1" \
+        : "=&r" (__temp), "=&r" (overflow) \
+        : "r" (__arg1), "r" (__arg2), "r" (__out), "g" (overflow), "0" ((ulong)0), "1" ((ulong)0)        : "cc"); \
+} while(0)
+
+#define mulll(a,b) \
+__extension__ ({ ulong __valuelo, __arg1 = (a), __arg2 = (b); \
+   __asm__ ("mull %3" \
+        : "=a" /* %eax */ (__valuelo), "=d" /* %edx */ (hiremainder) \
+        : "0" (__arg1), "rm" (__arg2)); \
+   __valuelo; \
+})
+
+#define addmul(a,b) \
+__extension__ ({ ulong __valuelo, __arg1 = (a), __arg2 = (b), __temp; \
+   __asm__ ("mull %4 ; addl %5,%0 ; adcl %6,%1" \
+        : "=a" /* %eax */ (__valuelo), "=&d" /* %edx */ (hiremainder), "=r" (__temp) \
+        : "0" (__arg1), "rm" (__arg2), "g" (hiremainder), "2" ((ulong)0)); \
+   __valuelo; \
+})
+
+#define divll(a,b) \
+__extension__ ({ ulong __value, __arg1 = (a), __arg2 = (b); \
+   __asm__ ("divl %4" \
+        : "=a" /* %eax */ (__value), "=d" /* %edx */ (hiremainder) \
+        : "0" /* %eax */ (__arg1), "1" /* %edx */ (hiremainder), "mr" (__arg2)); \
+   __value; \
+})
+
+#define bfffo(x) \
+__extension__ ({ ulong __arg = (x); \
+   int leading_one_position; \
+  __asm__ ("bsrl %1,%0" : "=r" (leading_one_position) : "rm" (__arg)); \
+  31 - leading_one_position; \
+})
+
 #endif
 #line 2 "../src/kernel/none/divll_pre.h"
 /* Copyright (C) 2014  The PARI group.
@@ -488,7 +611,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -498,7 +622,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #undef  LOCAL_HIREMAINDER
 extern ulong hiremainder;
 #if defined(INLINE) && defined(__GNUC__) && !defined(DISABLE_INLINE)
-#define LOCAL_HIREMAINDER register ulong hiremainder
+#define LOCAL_HIREMAINDER ulong hiremainder
 #else
 #define LOCAL_HIREMAINDER
 #endif
@@ -618,40 +742,38 @@ divll_pre(ulong a_lo, ulong n, ulong ninv)
 #define __AGM_ATAN_LIMIT                 60
 #define __DIVRR_GMP_LIMIT                4
 #define __EXPNEWTON_LIMIT                66
-#define __F2x_MUL_KARATSUBA_LIMIT        15
+#define __EXTGCD_HALFGCD_LIMIT           -1
+#define __F2x_MUL_KARATSUBA_LIMIT        11
+#define __F2x_MUL_MULII_LIMIT            41
 #define __F2xqX_BARRETT_LIMIT            48
-#define __F2xqX_DIVREM_BARRETT_LIMIT     292
-#define __F2xqX_INVBARRETT_LIMIT         46
-#define __F2xqX_REM_BARRETT_LIMIT        292
-#define __Flx_BARRETT_HALFMULII_LIMIT    21
-#define __Flx_BARRETT_KARATSUBA_LIMIT    1172
-#define __Flx_BARRETT_MULII2_LIMIT       16
-#define __Flx_BARRETT_MULII_LIMIT        448
-#define __Flx_BARRETT_QUARTMULII_LIMIT   23
-#define __Flx_DIVREM_BARRETT_LIMIT       768
-#define __Flx_EXTGCD_LIMIT               241
-#define __Flx_GCD_LIMIT                  1017
-#define __Flx_HALFGCD_HALFMULII_LIMIT    48
-#define __Flx_HALFGCD_KARATSUBA_LIMIT    77
-#define __Flx_HALFGCD_MULII2_LIMIT       25
-#define __Flx_HALFGCD_MULII_LIMIT        71
-#define __Flx_HALFGCD_QUARTMULII_LIMIT   52
-#define __Flx_INVBARRETT_HALFMULII_LIMIT 231
-#define __Flx_INVBARRETT_KARATSUBA_LIMIT 5067
-#define __Flx_INVBARRETT_MULII2_LIMIT    26
-#define __Flx_INVBARRETT_MULII_LIMIT     1154
-#define __Flx_INVBARRETT_QUARTMULII_LIMIT 139
-#define __Flx_MUL_HALFMULII_LIMIT        5
-#define __Flx_MUL_KARATSUBA_LIMIT        142
-#define __Flx_MUL_MULII2_LIMIT           5
-#define __Flx_MUL_MULII_LIMIT            7
-#define __Flx_MUL_QUARTMULII_LIMIT       5
-#define __Flx_REM_BARRETT_LIMIT          1266
-#define __Flx_SQR_HALFSQRI_LIMIT         3
-#define __Flx_SQR_KARATSUBA_LIMIT        316
-#define __Flx_SQR_QUARTSQRI_LIMIT        3
-#define __Flx_SQR_SQRI2_LIMIT            7
-#define __Flx_SQR_SQRI_LIMIT             5
+#define __F2xqX_DIVREM_BARRETT_LIMIT     97
+#define __F2xqX_EXTGCD_LIMIT             97
+#define __F2xqX_GCD_LIMIT                605
+#define __F2xqX_HALFGCD_LIMIT            127
+#define __F2xqX_INVBARRETT_LIMIT         50
+#define __F2xqX_REM_BARRETT_LIMIT        101
+#define __Flx_BARRETT2_LIMIT             31
+#define __Flx_BARRETT_LIMIT              92
+#define __Flx_DIVREM2_BARRETT_LIMIT      14
+#define __Flx_DIVREM_BARRETT_LIMIT       161
+#define __Flx_EXTGCD2_LIMIT              61
+#define __Flx_EXTGCD_LIMIT               158
+#define __Flx_GCD2_LIMIT                 1409
+#define __Flx_GCD_LIMIT                  852
+#define __Flx_HALFGCD2_LIMIT             36
+#define __Flx_HALFGCD_LIMIT              120
+#define __Flx_INVBARRETT2_LIMIT          22
+#define __Flx_INVBARRETT_LIMIT           200
+#define __Flx_MUL2_KARATSUBA_LIMIT       11
+#define __Flx_MUL2_MULII_LIMIT           8
+#define __Flx_MUL_KARATSUBA_LIMIT        33
+#define __Flx_MUL_MULII_LIMIT            30
+#define __Flx_REM2_BARRETT_LIMIT         89
+#define __Flx_REM_BARRETT_LIMIT          159
+#define __Flx_SQR2_KARATSUBA_LIMIT       15
+#define __Flx_SQR2_SQRI_LIMIT            14
+#define __Flx_SQR_KARATSUBA_LIMIT        93
+#define __Flx_SQR_SQRI_LIMIT             37
 #define __FlxqX_BARRETT_LIMIT            17
 #define __FlxqX_DIVREM_BARRETT_LIMIT     46
 #define __FlxqX_EXTGCD_LIMIT             44
@@ -675,6 +797,8 @@ divll_pre(ulong a_lo, ulong n, ulong ninv)
 #define __FpX_REM_BARRETT_LIMIT          111
 #define __Fp_POW_BARRETT_LIMIT           127
 #define __Fp_POW_REDC_LIMIT              17
+#define __GCD_HALFGCD_LIMIT              -1
+#define __HALFGCD_LIMIT                  12
 #define __INVMOD_GMP_LIMIT               3
 #define __INVNEWTON_LIMIT                75
 #define __LOGAGMCX_LIMIT                 22
@@ -691,40 +815,38 @@ divll_pre(ulong a_lo, ulong n, ulong ninv)
 #define __AGM_ATAN_LIMIT                 89
 #define __DIVRR_GMP_LIMIT                4
 #define __EXPNEWTON_LIMIT                197
-#define __F2x_MUL_KARATSUBA_LIMIT        23
-#define __F2xqX_BARRETT_LIMIT            17
-#define __F2xqX_DIVREM_BARRETT_LIMIT     46
-#define __F2xqX_INVBARRETT_LIMIT         22
-#define __F2xqX_REM_BARRETT_LIMIT        48
-#define __Flx_BARRETT_HALFMULII_LIMIT    23
-#define __Flx_BARRETT_KARATSUBA_LIMIT    905
-#define __Flx_BARRETT_MULII2_LIMIT       647
-#define __Flx_BARRETT_MULII_LIMIT        433
-#define __Flx_BARRETT_QUARTMULII_LIMIT   20
-#define __Flx_DIVREM_BARRETT_LIMIT       1289
-#define __Flx_EXTGCD_LIMIT               632
-#define __Flx_GCD_LIMIT                  2514
-#define __Flx_HALFGCD_HALFMULII_LIMIT    78
-#define __Flx_HALFGCD_KARATSUBA_LIMIT    139
-#define __Flx_HALFGCD_MULII2_LIMIT       537
-#define __Flx_HALFGCD_MULII_LIMIT        91
-#define __Flx_HALFGCD_QUARTMULII_LIMIT   37
-#define __Flx_INVBARRETT_HALFMULII_LIMIT 240
-#define __Flx_INVBARRETT_KARATSUBA_LIMIT 3600
-#define __Flx_INVBARRETT_MULII2_LIMIT    1815
-#define __Flx_INVBARRETT_MULII_LIMIT     1293
-#define __Flx_INVBARRETT_QUARTMULII_LIMIT 73
-#define __Flx_MUL_HALFMULII_LIMIT        7
-#define __Flx_MUL_KARATSUBA_LIMIT        90
-#define __Flx_MUL_MULII2_LIMIT           152
-#define __Flx_MUL_MULII_LIMIT            8
-#define __Flx_MUL_QUARTMULII_LIMIT       7
-#define __Flx_REM_BARRETT_LIMIT          689
-#define __Flx_SQR_HALFSQRI_LIMIT         4
-#define __Flx_SQR_KARATSUBA_LIMIT        159
-#define __Flx_SQR_QUARTSQRI_LIMIT        4
-#define __Flx_SQR_SQRI2_LIMIT            470
-#define __Flx_SQR_SQRI_LIMIT             5
+#define __EXTGCD_HALFGCD_LIMIT           -1
+#define __F2x_MUL_KARATSUBA_LIMIT        13
+#define __F2x_MUL_MULII_LIMIT            774
+#define __F2xqX_BARRETT_LIMIT            48
+#define __F2xqX_DIVREM_BARRETT_LIMIT     127
+#define __F2xqX_EXTGCD_LIMIT             127
+#define __F2xqX_GCD_LIMIT                884
+#define __F2xqX_HALFGCD_LIMIT            89
+#define __F2xqX_INVBARRETT_LIMIT         40
+#define __F2xqX_REM_BARRETT_LIMIT        127
+#define __Flx_BARRETT2_LIMIT             52
+#define __Flx_BARRETT_LIMIT              164
+#define __Flx_DIVREM2_BARRETT_LIMIT      111
+#define __Flx_DIVREM_BARRETT_LIMIT       470
+#define __Flx_EXTGCD2_LIMIT              184
+#define __Flx_EXTGCD_LIMIT               469
+#define __Flx_GCD2_LIMIT                 1281
+#define __Flx_GCD_LIMIT                  2817
+#define __Flx_HALFGCD2_LIMIT             181
+#define __Flx_HALFGCD_LIMIT              586
+#define __Flx_INVBARRETT2_LIMIT          397
+#define __Flx_INVBARRETT_LIMIT           501
+#define __Flx_MUL2_KARATSUBA_LIMIT       9
+#define __Flx_MUL2_MULII_LIMIT           8
+#define __Flx_MUL_KARATSUBA_LIMIT        57
+#define __Flx_MUL_MULII_LIMIT            146
+#define __Flx_REM2_BARRETT_LIMIT         89
+#define __Flx_REM_BARRETT_LIMIT          388
+#define __Flx_SQR2_KARATSUBA_LIMIT       18
+#define __Flx_SQR2_SQRI_LIMIT            20
+#define __Flx_SQR_KARATSUBA_LIMIT        112
+#define __Flx_SQR_SQRI_LIMIT             183
 #define __FlxqX_BARRETT_LIMIT            17
 #define __FlxqX_DIVREM_BARRETT_LIMIT     46
 #define __FlxqX_EXTGCD_LIMIT             44
@@ -748,6 +870,8 @@ divll_pre(ulong a_lo, ulong n, ulong ninv)
 #define __FpX_REM_BARRETT_LIMIT          127
 #define __Fp_POW_BARRETT_LIMIT           11
 #define __Fp_POW_REDC_LIMIT              3
+#define __GCD_HALFGCD_LIMIT              -1
+#define __HALFGCD_LIMIT                  22
 #define __INVMOD_GMP_LIMIT               3
 #define __INVNEWTON_LIMIT                93
 #define __LOGAGMCX_LIMIT                 32
@@ -768,7 +892,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -805,7 +930,8 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
@@ -898,6 +1024,26 @@ affrr_fixlg(GEN y, GEN z) { fixlg(z, lg(y)); affrr(y, z); }
 /*                       ALLOCATE ON STACK                         */
 /*                                                                 */
 /*******************************************************************/
+INLINE ulong
+get_avma(void) { return avma; }
+INLINE void
+set_avma(ulong av) { avma = av; }
+
+INLINE double
+gc_double(pari_sp av, double d) { set_avma(av); return d; }
+INLINE long
+gc_long(pari_sp av, long s) { set_avma(av); return s; }
+INLINE ulong
+gc_ulong(pari_sp av, ulong s) { set_avma(av); return s; }
+INLINE int
+gc_bool(pari_sp av, int s) { set_avma(av); return s; }
+INLINE int
+gc_int(pari_sp av, int s) { set_avma(av); return s; }
+INLINE GEN
+gc_NULL(pari_sp av) { set_avma(av); return NULL; }
+INLINE GEN
+gc_const(pari_sp av, GEN x) { set_avma(av); return x; }
+
 INLINE GEN
 new_chunk(size_t x) /* x is a number of longs */
 {
@@ -905,7 +1051,7 @@ new_chunk(size_t x) /* x is a number of longs */
   CHECK_CTRLC
   if (x > (avma-pari_mainstack->bot) / sizeof(long))
     new_chunk_resize(x);
-  avma = (pari_sp)z;
+  set_avma((pari_sp)z);
 #ifdef MEMSTEP
   if (DEBUGMEM>1 && pari_mainstack->memused != DISABLE_MEMUSED) {
     long d = (long)pari_mainstack->memused - (long)z;
@@ -930,8 +1076,9 @@ stack_malloc(size_t N)
 INLINE char *
 stack_malloc_align(size_t N, long k)
 {
-  ulong d = ((ulong)avma) % k;
+  ulong d = ((ulong)avma) % k, e = ((ulong)N) % k;
   if (d) (void)new_chunk(d/sizeof(long));
+  if (e) N += k-e;
   return (char*) new_chunk(nchar2nlong(N));
 }
 
@@ -942,6 +1089,15 @@ stack_calloc(size_t N)
   memset(p, 0, N); return p;
 }
 
+INLINE char *
+stack_calloc_align(size_t N, long k)
+{
+  ulong d = ((ulong)avma) % k, e = ((ulong)N) % k;
+  if (d) (void)new_chunk(d/sizeof(long));
+  if (e) N += k-e;
+  return stack_calloc(N);
+}
+
 /* cgetg(lg(x), typ(x)), set *lx. Implicit unsetisclone() */
 INLINE GEN
 cgetg_copy(GEN x, long *plx) {
@@ -1008,7 +1164,7 @@ cgetr(long x)
 INLINE GEN
 leafcopy(GEN x)
 {
-  register long lx = lg(x);
+  long lx = lg(x);
   GEN y = new_chunk(lx); /* can't use cgetg_copy, in case x,y overlap */
   while (--lx > 0) y[lx] = x[lx];
   y[0] = x[0] & (TYPBITS|LGBITS); return y;
@@ -1047,8 +1203,6 @@ INLINE GEN
 mpneg(GEN x) { GEN y = leafcopy(x); togglesign(y); return y; }
 INLINE GEN negi(GEN x) { return mpneg(x); }
 INLINE GEN negr(GEN x) { return mpneg(x); }
-INLINE GEN gmax_shallow(GEN x, GEN y) { return gcmp(x,y)<0? y: x; }
-INLINE GEN gmin_shallow(GEN x, GEN y) { return gcmp(x,y)<0? x: y; }
 
 /* negate in place */
 INLINE void
@@ -1225,6 +1379,14 @@ INLINE ulong int_bit(GEN x, long n)
 /*                           COMPARISON                            */
 /*                                                                 */
 /*******************************************************************/
+INLINE int
+cmpss(long a, long b)
+{ return a>b? 1: (a<b? -1: 0); }
+
+INLINE int
+cmpuu(ulong a, ulong b)
+{ return a>b? 1: (a<b? -1: 0); }
+
 INLINE int
 cmpir(GEN x, GEN y)
 {
@@ -1237,7 +1399,7 @@ cmpir(GEN x, GEN y)
     if (expo(y) >= expi(x)) return 0;
     return signe(x);
   }
-  av=avma; z = itor(x, realprec(y)); avma=av;
+  av=avma; z = itor(x, realprec(y)); set_avma(av);
   return cmprr(z,y); /* cmprr does no memory adjustment */
 }
 INLINE int
@@ -1249,7 +1411,7 @@ cmpsr(long x, GEN y)
   GEN z;
 
   if (!x) return -signe(y);
-  av=avma; z = stor(x, LOWDEFAULTPREC); avma=av;
+  av=avma; z = stor(x, LOWDEFAULTPREC); set_avma(av);
   return cmprr(z,y);
 }
 INLINE int
@@ -1481,7 +1643,7 @@ remii(GEN a, GEN b) { return dvmdii(a,b,ONLY_REM); }
 INLINE GEN
 divss(long x, long y) { return stoi(x / y); }
 INLINE GEN
-modss(long x, long y) { return stoi(smodss(x, y)); }
+modss(long x, long y) { return utoi(smodss(x, y)); }
 INLINE GEN
 remss(long x, long y) { return stoi(x % y); }
 INLINE long
@@ -1518,8 +1680,9 @@ udivuu_rem(ulong x, ulong y, ulong *r)
 INLINE ulong
 ceildivuu(ulong a, ulong b)
 {
-  ulong c = a/b;
-  return (a%b)? c+1: c;
+  ulong c;
+  if (!a) return 0;
+  c = a / b; return (a % b)? c+1: c;
 }
 
 INLINE ulong
@@ -1620,15 +1783,14 @@ INLINE long
 smodis(GEN x, long y)
 {
   pari_sp av = avma;
-  long r;
-  (void)divis_rem(x,y, &r); avma = av; return (r >= 0) ? r: labs(y) + r;
+  long r; (void)divis_rem(x,y, &r);
+  return gc_long(av, (r >= 0)? r: labs(y) + r);
 }
 INLINE GEN
 modis(GEN x, long y) { return stoi(smodis(x,y)); }
 INLINE GEN
 modsi(long x, GEN y) {
-  long r;
-  (void)sdivsi_rem(x, y, &r);
+  long r; (void)sdivsi_rem(x, y, &r);
   return (r >= 0)? stoi(r): addsi_sign(r, y, 1);
 }
 
@@ -1653,7 +1815,7 @@ remis(GEN x, long y)
 {
   pari_sp av = avma;
   long r;
-  (void)divis_rem(x,y, &r); avma = av; return stoi(r);
+  (void)divis_rem(x,y, &r); set_avma(av); return stoi(r);
 }
 
 INLINE GEN
@@ -1662,7 +1824,7 @@ rdivis(GEN x, long y, long prec)
   GEN z = cgetr(prec);
   pari_sp av = avma;
   affrr(divrs(itor(x,prec), y),z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 INLINE GEN
 rdivsi(long x, GEN y, long prec)
@@ -1670,7 +1832,7 @@ rdivsi(long x, GEN y, long prec)
   GEN z = cgetr(prec);
   pari_sp av = avma;
   affrr(divsr(x, itor(y,prec)), z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 INLINE GEN
 rdivss(long x, long y, long prec)
@@ -1678,57 +1840,61 @@ rdivss(long x, long y, long prec)
   GEN z = cgetr(prec);
   pari_sp av = avma;
   affrr(divrs(stor(x, prec), y), z);
-  avma = av; return z;
+  set_avma(av); return z;
 }
 
 INLINE void
 rdiviiz(GEN x, GEN y, GEN z)
 {
-  pari_sp av = avma;
-  long prec = realprec(z);
-  affir(x, z);
-  if (!is_bigint(y)) {
-    affrr(divrs(z, y[2]), z);
-    if (signe(y) < 0) togglesign(z);
+  long prec = realprec(z), lx = lgefint(x), ly = lgefint(y);
+  if (lx == 2) { affur(0, z); return; }
+  if (ly == 3)
+  {
+    affir(x, z); if (signe(y) < 0) togglesign(z);
+    affrr(divru(z, y[2]), z);
+  }
+  else if (lx > prec + 1 || ly > prec + 1)
+  {
+    affir(x,z); affrr(divri(z, y), z);
   }
   else
-    affrr(divrr(z, itor(y,prec)), z);
-  avma = av;
+  {
+    long b = bit_accuracy(prec) + expi(y) - expi(x) + 1;
+    GEN q = divii(b > 0? shifti(x, b): x, y);
+    affir(q, z); if (b > 0) shiftr_inplace(z, -b);
+  }
+  set_avma((ulong)z);
 }
 INLINE GEN
 rdivii(GEN x, GEN y, long prec)
-{
-  GEN z = itor(x, prec);
-  pari_sp av = avma;
-  if (lg(y) == 3) {
-    affrr(divru(z, y[2]), z);
-    if (signe(y) < 0) togglesign(z);
-  }
-  else
-    affrr(divrr(z, itor(y,prec)), z);
-  avma = av; return z;
-}
+{ GEN z = cgetr(prec); rdiviiz(x, y, z); return z; }
 INLINE GEN
-fractor(GEN x, long prec) { return rdivii(gel(x,1), gel(x,2), prec); }
+fractor(GEN x, long prec)
+{ return rdivii(gel(x,1), gel(x,2), prec); }
 
 INLINE int
 dvdii(GEN x, GEN y)
 {
-  pari_sp av=avma;
-  GEN r = remii(x,y);
-  avma = av; return r == gen_0;
+  pari_sp av = avma;
+  GEN r;
+  if (!signe(x)) return 1;
+  if (!signe(y)) return 0;
+  r = remii(x,y);
+  return gc_bool(av, r == gen_0);
 }
 INLINE int
 dvdsi(long x, GEN y)
 {
-  if (!signe(y)) return x == 0;
+  if (x == 0) return 1;
+  if (!signe(y)) return 0;
   if (lgefint(y) != 3) return 0;
   return x % y[2] == 0;
 }
 INLINE int
 dvdui(ulong x, GEN y)
 {
-  if (!signe(y)) return x == 0;
+  if (x == 0) return 1;
+  if (!signe(y)) return 0;
   if (lgefint(y) != 3) return 0;
   return x % y[2] == 0;
 }
@@ -1745,7 +1911,7 @@ dvdisz(GEN x, long y, GEN z)
   const pari_sp av = avma;
   long r;
   GEN p1 = divis_rem(x,y, &r);
-  avma = av; if (r) return 0;
+  set_avma(av); if (r) return 0;
   affii(p1,z); return 1;
 }
 INLINE int
@@ -1754,18 +1920,16 @@ dvdiuz(GEN x, ulong y, GEN z)
   const pari_sp av = avma;
   ulong r;
   GEN p1 = absdiviu_rem(x,y, &r);
-  avma = av; if (r) return 0;
+  set_avma(av); if (r) return 0;
   affii(p1,z); return 1;
 }
 INLINE int
 dvdiiz(GEN x, GEN y, GEN z)
 {
   const pari_sp av=avma;
-  GEN p2;
-  const GEN p1=dvmdii(x,y,&p2);
-
-  if (signe(p2)) { avma=av; return 0; }
-  affii(p1,z); avma=av; return 1;
+  GEN p2, p1 = dvmdii(x,y,&p2);
+  if (signe(p2)) return gc_bool(av,0);
+  affii(p1,z); return gc_bool(av,1);
 }
 
 INLINE ulong
@@ -1778,7 +1942,7 @@ remlll_pre(ulong u2, ulong u1, ulong u0, ulong n, ulong ninv)
 INLINE ulong
 Fl_sqr_pre(ulong a, ulong p, ulong pi)
 {
-  register ulong x;
+  ulong x;
   LOCAL_HIREMAINDER;
   x = mulll(a,a);
   return remll_pre(hiremainder, x, p, pi);
@@ -1787,7 +1951,7 @@ Fl_sqr_pre(ulong a, ulong p, ulong pi)
 INLINE ulong
 Fl_mul_pre(ulong a, ulong b, ulong p, ulong pi)
 {
-  register ulong x;
+  ulong x;
   LOCAL_HIREMAINDER;
   x = mulll(a,b);
   return remll_pre(hiremainder, x, p, pi);
@@ -1927,7 +2091,7 @@ Fl_center(ulong u, ulong p, ulong ps2) { return (long) (u > ps2)? u - p: u; }
 INLINE ulong
 Fl_mul(ulong a, ulong b, ulong p)
 {
-  register ulong x;
+  ulong x;
   LOCAL_HIREMAINDER;
   x = mulll(a,b);
   if (!hiremainder) return x % p;
@@ -1936,14 +2100,16 @@ Fl_mul(ulong a, ulong b, ulong p)
 INLINE ulong
 Fl_sqr(ulong a, ulong p)
 {
-  register ulong x;
+  ulong x;
   LOCAL_HIREMAINDER;
   x = mulll(a,a);
   if (!hiremainder) return x % p;
   (void)divll(x,p); return hiremainder;
 }
+/* don't assume that p is prime: can't special case a = 0 */
 INLINE ulong
-Fl_div(ulong a, ulong b, ulong p) { return Fl_mul(a, Fl_inv(b, p), p); }
+Fl_div(ulong a, ulong b, ulong p)
+{ return Fl_mul(a, Fl_inv(b, p), p); }
 
 /*******************************************************************/
 /*                                                                 */
@@ -1994,7 +2160,6 @@ vali(GEN x)
   return vals(*xp) + i * BITS_IN_LONG;
 }
 
-
 /* assume x > 0 */
 INLINE long
 expu(ulong x) { return (BITS_IN_LONG-1) - (long)bfffo(x); }
@@ -2144,116 +2309,116 @@ mpaff(GEN x, GEN y) { if (typ(x)==t_INT) affiz(x, y); else affrr(x,y); }
 /*******************************************************************/
 
 INLINE void addiiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(addii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(addii(x,y),z); set_avma(av); }
 INLINE void addirz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(addir(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addir(x,y),z); set_avma(av); }
 INLINE void addriz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(addri(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addri(x,y),z); set_avma(av); }
 INLINE void addrrz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(addrr(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addrr(x,y),z); set_avma(av); }
 INLINE void addsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(addsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(addsi(s,y),z); set_avma(av); }
 INLINE void addsrz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affrr(addsr(s,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addsr(s,y),z); set_avma(av); }
 INLINE void addssz(long s, long y, GEN z)
-{ pari_sp av = avma; affii(addss(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(addss(s,y),z); set_avma(av); }
 
 INLINE void diviiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(divii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(divii(x,y),z); set_avma(av); }
 INLINE void divirz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(divir(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(divir(x,y),z); set_avma(av); }
 INLINE void divisz(GEN x, long y, GEN z)
-{ pari_sp av = avma; affii(divis(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(divis(x,y),z); set_avma(av); }
 INLINE void divriz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(divri(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(divri(x,y),z); set_avma(av); }
 INLINE void divrrz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(divrr(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(divrr(x,y),z); set_avma(av); }
 INLINE void divrsz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affrr(divrs(y,s),z); avma = av; }
+{ pari_sp av = avma; affrr(divrs(y,s),z); set_avma(av); }
 INLINE void divsiz(long x, GEN y, GEN z)
 { long junk; affsi(sdivsi_rem(x,y,&junk), z); }
 INLINE void divsrz(long s, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(divsr(s,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(divsr(s,y),z); set_avma(av); }
 INLINE void divssz(long x, long y, GEN z)
 { affsi(x/y, z); }
 
 INLINE void modisz(GEN y, long s, GEN z)
 { affsi(smodis(y,s),z); }
 INLINE void modsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(modsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(modsi(s,y),z); set_avma(av); }
 INLINE void modssz(long s, long y, GEN z)
 { affsi(smodss(s,y),z); }
 
 INLINE void mpaddz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mpadd(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mpadd(x,y),z); set_avma(av); }
 INLINE void mpsubz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mpsub(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mpsub(x,y),z); set_avma(av); }
 INLINE void mpmulz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mpmul(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mpmul(x,y),z); set_avma(av); }
 
 INLINE void muliiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(mulii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(mulii(x,y),z); set_avma(av); }
 INLINE void mulirz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mulir(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mulir(x,y),z); set_avma(av); }
 INLINE void mulriz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mulri(x,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mulri(x,y),z); set_avma(av); }
 INLINE void mulrrz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(mulrr(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(mulrr(x,y),z); set_avma(av); }
 INLINE void mulsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(mulsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(mulsi(s,y),z); set_avma(av); }
 INLINE void mulsrz(long s, GEN y, GEN z)
-{ pari_sp av = avma; mpaff(mulsr(s,y),z); avma = av; }
+{ pari_sp av = avma; mpaff(mulsr(s,y),z); set_avma(av); }
 INLINE void mulssz(long s, long y, GEN z)
-{ pari_sp av = avma; affii(mulss(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(mulss(s,y),z); set_avma(av); }
 
 INLINE void remiiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(remii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(remii(x,y),z); set_avma(av); }
 INLINE void remisz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affii(remis(y,s),z); avma = av; }
+{ pari_sp av = avma; affii(remis(y,s),z); set_avma(av); }
 INLINE void remsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(remsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(remsi(s,y),z); set_avma(av); }
 INLINE void remssz(long s, long y, GEN z)
-{ pari_sp av = avma; affii(remss(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(remss(s,y),z); set_avma(av); }
 
 INLINE void subiiz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affii(subii(x,y),z); avma = av; }
+{ pari_sp av = avma; affii(subii(x,y),z); set_avma(av); }
 INLINE void subirz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(subir(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(subir(x,y),z); set_avma(av); }
 INLINE void subisz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affii(addsi(-s,y),z); avma = av; }
+{ pari_sp av = avma; affii(addsi(-s,y),z); set_avma(av); }
 INLINE void subriz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(subri(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(subri(x,y),z); set_avma(av); }
 INLINE void subrrz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; affrr(subrr(x,y),z); avma = av; }
+{ pari_sp av = avma; affrr(subrr(x,y),z); set_avma(av); }
 INLINE void subrsz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affrr(addsr(-s,y),z); avma = av; }
+{ pari_sp av = avma; affrr(addsr(-s,y),z); set_avma(av); }
 INLINE void subsiz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affii(subsi(s,y),z); avma = av; }
+{ pari_sp av = avma; affii(subsi(s,y),z); set_avma(av); }
 INLINE void subsrz(long s, GEN y, GEN z)
-{ pari_sp av = avma; affrr(subsr(s,y),z); avma = av; }
+{ pari_sp av = avma; affrr(subsr(s,y),z); set_avma(av); }
 INLINE void subssz(long x, long y, GEN z) { addssz(x,-y,z); }
 
 INLINE void
 dvmdssz(long x, long y, GEN z, GEN t) {
   pari_sp av = avma;
   long r;
-  affii(divss_rem(x,y, &r), z); avma = av; affsi(r,t);
+  affii(divss_rem(x,y, &r), z); set_avma(av); affsi(r,t);
 }
 INLINE void
 dvmdsiz(long x, GEN y, GEN z, GEN t) {
   pari_sp av = avma;
   long r;
-  affii(divsi_rem(x,y, &r), z); avma = av; affsi(r,t);
+  affii(divsi_rem(x,y, &r), z); set_avma(av); affsi(r,t);
 }
 INLINE void
 dvmdisz(GEN x, long y, GEN z, GEN t) {
   pari_sp av = avma;
   long r;
-  affii(divis_rem(x,y, &r),z); avma = av; affsi(r,t);
+  affii(divis_rem(x,y, &r),z); set_avma(av); affsi(r,t);
 }
 INLINE void
 dvmdiiz(GEN x, GEN y, GEN z, GEN t) {
   pari_sp av = avma;
   GEN r;
-  affii(dvmdii(x,y,&r),z); affii(r,t); avma=av;
+  affii(dvmdii(x,y,&r),z); affii(r,t); set_avma(av);
 }
