@@ -7,17 +7,16 @@
 #define GPDATADIR "/usr/local/share/pari"
 #define SHELL_Q '\''
 
-#define PARIVERSION "GP/PARI CALCULATOR Version 2.11.4 (released)"
-#define PARIINFO "riscv64 running openbsd (portable C/GMP-6.2.1 kernel) 64-bit version"
-#define PARI_VERSION_CODE 133892
+#define PARIVERSION "GP/PARI CALCULATOR Version 2.15.4 (released)"
+#define PARIINFO "i386 running openbsd (ix86/GMP-%s kernel) 32-bit version"
+#define PARI_VERSION_CODE 134916
 #define PARI_VERSION(a,b,c) (((a) << 16) + ((b) << 8) + (c))
 #define PARI_VERSION_SHIFT 8
 #define PARI_VCSVERSION ""
 #define PARI_MT_ENGINE "single"
 
-#define PARI_DOUBLE_FORMAT -
+#define PARI_DOUBLE_FORMAT 1
 #define GCC_VERSION "OpenBSD clang version 13.0.0"
-#define __HAS_NO_ASM__
 #define ASMINLINE
 
 /*  Location of GNU gzip program (enables reading of .Z and .gz files). */
@@ -28,9 +27,9 @@
 #define GP_MIME_OPEN "xdg-open"
 
 #define READLINE "4.3"
-#define LONG_IS_64BIT
 #define HAS_EXP2
 #define HAS_LOG2
+#define HAS_RINT
 #define HAS_ISATTY
 #define HAS_ALARM
 #define HAS_SYSTEM
